/* Classic Ladder Project */
/* Copyright (C) 2001-2005 Marc Le Douarain */
/* http://www.multimania.com/mavati/classicladder */
/* http://www.sourceforge.net/projects/classicladder */
/* August 2005 */
/* ------------------------------------ */
/* Serial low-level functions for Linux */
/* ------------------------------------ */

/* This library is free software; you can redistribute it and/or */
/* modify it under the terms of the GNU Lesser General Public */
/* License as published by the Free Software Foundation; either */
/* version 2.1 of the License, or (at your option) any later version. */

/* This library is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/* Lesser General Public License for more details. */

/* You should have received a copy of the GNU Lesser General Public */
/* License along with this library; if not, write to the Free Software */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/* Modified by Dave Cole 5/24/10 to increase serial receive reliability, 
   correct errors in EMC2 implementation version*/

#include <termios.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h> 
#include <sys/ioctl.h>
#include <errno.h>       //  Added 

#include "serial_common.h"
extern int ModbusSerialUseRtsToSend;
extern int ModbusDebugLevel;
extern int ModbusSerialDataBits;   // Number of data bits (7, 6, 7, 8)
extern int ModbusSerialStopBits;   // Number of stop bits (1 or 2)
extern int ModbusSerialParity;      // Parity (00 = NONE, 01 = Odd, 02 = Even, 03 = Mark, 04 = Space)

int SerialSpeed[ ] = { /*300,*/ 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 0 };  // remove 300 baud
int SerialCorres[ ] = { /*B300,*/ B600, B1200, B2400, B4800, B9600, B19200, B38400, B57600, B115200, 0 };

char PortIsOpened = 0;
int fd;
long DATABITS;
long STOPBITS;
long PARITYON;
long PARITY;



struct termios oldtio;
struct termios newtio;

char SerialOpen( char * SerialPortName, int Speed )
{
	/* if port already opened => close it before */
	if ( PortIsOpened )
		SerialClose( );

	/* open the device to be non-blocking (read will return immediatly) */
	fd = open( SerialPortName, O_RDWR | O_NOCTTY | O_NDELAY/*don't wait DTR*/ );
        //fd = open( SerialPortName, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd >=0)
	{
		int BaudRate = -1;
		int ScanBaudRate = 0;
		while( SerialSpeed[ ScanBaudRate ]!=Speed && SerialSpeed[ ScanBaudRate ]>=0 )
		{
			ScanBaudRate++;
			if ( SerialSpeed[ ScanBaudRate ]==Speed )
			{
				BaudRate = SerialCorres[ ScanBaudRate ];
			}
		}
switch (ModbusSerialDataBits)
      {
         case 8:
         default:
            DATABITS = CS8;
            break;
         case 7:
            DATABITS = CS7;
            break;
         case 6:
            DATABITS = CS6;
            break;
         case 5:
            DATABITS = CS5;
            break;
      }  //end of switch data_bits
      switch (ModbusSerialStopBits)
      {
         case 1:
         default:
            STOPBITS = 0;
            break;
         case 2:
            STOPBITS = CSTOPB;
            break;
      }  //end of switch stop bits
      switch (ModbusSerialParity)
      {
         case 0:
         default:                       //none
            PARITYON = 0;
            PARITY = 0;
            break;
         case 1:                        //odd
            PARITYON = PARENB;
            PARITY = PARODD;
            break;
         case 2:                        //even
            PARITYON = PARENB;
            PARITY = 0;
            break;
      }  //end of switch parity
		if ( BaudRate!=-1 )
		{        
			tcgetattr(fd,&oldtio); /* save current port settings */
			/* set new port settings */
			bzero(&newtio, sizeof(newtio));
			//newtio.c_cflag = BaudRate | /*CRTSCTS |*/ CS8 | CLOCAL | CREAD;
                        newtio.c_cflag = BaudRate | DATABITS | STOPBITS | PARITYON | PARITY | CLOCAL | CREAD;
			//newtio.c_cflag |= PARENB
			newtio.c_iflag = IGNPAR    | IGNBRK; // | ICRNL;
			newtio.c_oflag = 0;
			newtio.c_lflag = 0;
			newtio.c_cc[VMIN]=0; //1;
			newtio.c_cc[VTIME]=0;
			tcsetattr(fd,TCSANOW,&newtio);
			PortIsOpened = 1;
		}
		else
		{
			printf("Speed value not found for serial\n");
		}
	}
	else
	{
		printf( "Failed to open serial port %s...\n", SerialPortName );
	}
	return PortIsOpened;
}

void SerialClose( )
{
	if ( PortIsOpened )
	{
		PortIsOpened = 0;
		/* restore old port settings */
		tcsetattr(fd,TCSANOW,&oldtio);
		close(fd);
	}
}

void SerialSetRTS( int State )
{
     int status;
	if ( PortIsOpened )
	{
		ioctl(fd, TIOCMGET, &status);
		if ( State )
			status |= TIOCM_RTS;
		else
			status &= ~TIOCM_RTS;
		if ( ModbusDebugLevel>=3 )
			printf( "Set RTS=%d\n", State );
		ioctl(fd, TIOCMSET, &status);
	}
}

void SerialSend( char *Buff, int BuffLength )
{
	if ( PortIsOpened )
	{
		if ( ModbusSerialUseRtsToSend )
		{
			SerialSetRTS( 1 );
		}
		if ( ModbusDebugLevel>=2 )
			printf("Serial writing...\n");

		write(fd,Buff,BuffLength);

		if ( ModbusDebugLevel>=2 )
			printf("Writing done!\n");
		if ( ModbusSerialUseRtsToSend )
		{
			// wait until everything has been transmitted
			tcdrain( fd );
			SerialSetRTS( 0 );
		}
	}
}

void SerialSetResponseSize( int Size, int TimeOutResp )
{
	if ( PortIsOpened )
	{         // "fd" (file device) control for serial port per termios docs 
	    newtio.c_cc[VMIN] = Size; //Nbr chars we should receive  (This doesn't seem to work, read returns immediately, regardless)
		newtio.c_cc[VTIME] = TimeOutResp/100; // TimeOut in 0.1s (Read returns immediately and doesn't block as it should? )
		if ( ModbusDebugLevel>=2 )
			printf("Serial config...\n");
		tcsetattr(fd,TCSANOW,&newtio);// TCSANOW - means make changes immediately
	}
}



int SerialReceive( char * Buff, int framelength , int TimeOutResp )
{

	int select_ret;  //added 
	int read_ret;
	char *p_response;
	int size_to_read;
	int num_rec;
	if ( ModbusDebugLevel>=2 )
		printf("Start of SerialReceive() \n");

	if ( PortIsOpened )
	{

		fd_set myset;  // does a struct type function to create myset ??
		struct timeval tv; // make a timer struct called tv

		if ( ModbusDebugLevel>=3 )
			printf("First select() \n"); // checks on the status of the file "fd"

		//  Wait for data routine 		
		FD_ZERO( &myset);  // clears a set
		FD_SET( fd, &myset ); // adds a file descriptor to myset called fd
		tv.tv_sec = 0; //seconds
		tv.tv_usec = TimeOutResp *1000; //micro-seconds				
		while ((select_ret = select(fd+1, &myset, NULL, NULL, &tv)) == -1) 
		{	
			if (errno == EINTR) 
				{					

				if ( ModbusDebugLevel>=2 )
					printf("A non blocked signal was caught in wait 1\n");		
									
				FD_ZERO(&myset);			//Necessary after an error 				
				FD_SET(fd, &myset);	

				if ( ModbusDebugLevel>=2 )	
					printf("errno == EINTR in wait 1\n");			
				} 	
		}										
			if (select_ret == 0)
				{
				if ( ModbusDebugLevel>=2 )
					printf("Comm timeout on select - in wait 1\n");	
				return	0;
				}						
		// end of wait for data routine   

		num_rec = 0;
		size_to_read = framelength;
		p_response = Buff;

	  while (select_ret)    // select_ret is >1 or true if select() has found chars are in buffer
		{
			if ( ModbusDebugLevel>=2 )
			  printf("size to read = %i\n",size_to_read);

			  read_ret = read(fd, p_response, size_to_read);

			if (read_ret == -1) 
				{     // read failed 
					if ( ModbusDebugLevel>=2 )
 						 printf("Read Port failure\n");	
				  return 0;
				} 
			else 
				{    // read was good
				  
				  num_rec += read_ret;  //Sums bytes received

				if ( ModbusDebugLevel>=2 )
					printf("total chars read = %i, chars read this time = %i\n", num_rec, read_ret);

					p_response = &(p_response[read_ret]);  // Moves the pointer to receive other datas
					size_to_read = framelength - num_rec; 

				}

			if (size_to_read > 0) 
				{

					if ( ModbusDebugLevel>=2 )	
						printf("More data to get during this read function\n");
			
					//  Wait for data routine again if not all of it was received during the first read 		
					FD_ZERO( &myset);  // clears a set
					FD_SET( fd, &myset ); // adds a file descriptor to myset called fd
					tv.tv_sec = 0; //seconds
					tv.tv_usec = TimeOutResp *1000; //micro-seconds		

					while ((select_ret = select(fd+1, &myset, NULL, NULL, &tv)) == -1) 
					{	
						if (errno == EINTR) 
						{					

							if ( ModbusDebugLevel>=2 )
								printf("A non blocked signal was caught in wait 2 \n");		

						// Necessary after an error 					
						FD_ZERO(&myset);							
						FD_SET(fd, &myset);	

							if ( ModbusDebugLevel>=2 )
								printf("errno == EINTR in wait 2\n");	
						} 
					}										
	
					if (select_ret == 0)
						{
						if ( ModbusDebugLevel>=2 )
						printf("Comm timeout on select in wait 2\n");	
						return	0;
						}	
					// end of wait for data routine

				} 
				else 
				{
				  /* All chars are received */
				  select_ret = 0;
				}
		}

		  /* OK */
		  return framelength;
	}
}



void SerialFlush( void )
{
	if ( PortIsOpened )
	{
		if ( ModbusDebugLevel>=2 )
			printf("Serial flush all!\n");
		tcflush( fd, TCIOFLUSH );
	//	usleep( 250*1000 );         // doing the flush twice with a time delay in between
	//	tcflush( fd, TCIOFLUSH );
	}
}

